#!/usr/bin/env python3

import sqlite3
import json
import sys
import os
import time
import ssl
import urllib.request
import base64
import db_schema
import hashlib
import threading

from logger import log
from email import encoders
from email.mime.base import MIMEBase
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from smtplib import SMTP, SMTPException, SMTPServerDisconnected
from datetime import datetime, timedelta, timezone
from configparser import ConfigParser
from signal import signal, SIGTERM
from pathlib import Path
from tempfile import gettempdir

class Orchid:
    __slots__ = ('uuid', 'name', 'proxy', 'tz')

    def __init__(self, orchid_info):
        self.uuid = orchid_info['id']
        self.name = orchid_info['name']
        self.proxy = ('{uri}/service/orchid-proxy/{uuid}'
                      .format(uri=fusion_uri, uuid=orchid_info['id']))
        self.tz = self._get_tz()

    def local_datetime(self, timestamp_ms=None):
        """Obtain Orchid's local datetime, relative to its timezone offset.

        If 'timestamp_ms' is specified return the local datetime associated to
        that timestamp, otherwise return the current local datetime.
        """
        if not self.tz:
            return None
        # break Orchid's offset into hours, minutes
        hrs, mins = divmod(self.tz['offset'], 60)
        td = timedelta(hours=hrs, minutes=mins)
        # apply the time difference to UTC datetime
        if timestamp_ms:
            return utc_datetime(timestamp=timestamp_ms/1000) + td
        return utc_datetime() + td

    def _get_tz(self):
        """Internal: obtain Orchid's timezone information (name, offset).
        """
        try:
            time_info = request('{}/service/time-extended'.format(self.proxy))
            return {
                'offset': time_info['tzOffset'],
                'name': time_info['tz']
            }
        except urllib.error.HTTPError as http_err:
            log.error('Failed to obtain timezone information for %s '
                      '(HTTP %d)', self.name, http_err.code)
            return None
        except OSError as os_err:
            log.error('Unable to request timezone information for %s: %s',
                      self.name, os_err)
            return None

class Schedule:
    __slots__ = ('start', 'stop', 'days')

    day_of_week = ('monday', 'tuesday', 'wednesday',
            'thursday', 'friday', 'saturday', 'sunday')

    def __init__(self, start, stop, days):
        # using 'HH:MM' 24-hour format
        self.start = '{:02d}:{:02d}'.format(*divmod(start, 60))
        self.stop = '{:02d}:{:02d}'.format(*divmod(stop, 60))
        self.days = set(days.split(','))

class MotionAlerter:
    __slots__ = ('orchid', 'recipients', 'schedules', 'cameras')

    def __init__(self, orchid_info, recipients):
        self.orchid = Orchid(orchid_info)
        self.recipients = recipients.split(',')
        self.schedules = self._load_schedules()
        self.cameras = self._load_cameras(orchid_info['cameras'])

    def in_schedule(self):
        """Check if MotionAlerter is within an active alert schedule.

        Schedule start-stop times may represent a period during the day
        (e.g. 10:00am - 6:30pm), or overnight (e.g. 10:00pm - 6:30am).
        """
        orchid_datetime = self.orchid.local_datetime()
        if not orchid_datetime:
            return False
        today = Schedule.day_of_week[orchid_datetime.weekday()]
        orchid_local_time = orchid_datetime.strftime('%H:%M')
        for schedule in self.schedules:
            if today not in schedule.days:
                continue
            if schedule.stop > schedule.start:
                if (orchid_local_time >= schedule.start
                and orchid_local_time <= schedule.stop):
                    return True
            else:
                if (orchid_local_time >= schedule.start
                or  orchid_local_time <= schedule.stop):
                    return True
        return False

    def alert_if_motion_detected(self):
        """Detect motion events per camera and perform async export/email.

        For ongoing events (start == stop) we set an instantaneous duration
        with the current UTC time based on whatever the start time is.

        Any duration less than 'min_motion_ms' is skipped. Any duration over
        'max_export_ms' is capped to that max. If an event has already been
        accounted for (recorded in the alert table), it is skipped.
        """

        THIRTY_SEC = 30 * 1000
        FIFTEEN_MIN = 15 * 60*1000

        min_motion_ms = config['motion'].getint('min_motion_sec', 5) * 1000
        max_export_ms = config['export'].getint('max_export_sec', 30) * 1000
        email_deadband_ms = config['smtp'].getint('email_deadband_sec', 900) * 1000

        jobs = []
        for camera_id in self.cameras:
            # obtain motion events for current camera_id
            utc_time_ms = int(utc_datetime(epoch=True) * 1000)
            last_check_ms = db.execute('''
            SELECT last_check
              FROM orchid_camera
             WHERE orchid_uuid=? AND camera_id=?
               ''', (self.orchid.uuid, camera_id)).fetchone()
            # if the user removed this cam between when it was
            # loaded and now, just skip and continue
            if not last_check_ms:
                continue
            last_check_ms, = last_check_ms
            try:
                db_lock.acquire()
                with db:
                    db.execute('''
                    UPDATE orchid_camera
                       SET last_check=?
                     WHERE orchid_uuid=? AND camera_id=?
                       ''', (utc_time_ms, self.orchid.uuid, camera_id))
            finally: db_lock.release()
            if utc_time_ms - last_check_ms > FIFTEEN_MIN:
                event_start_ms = utc_time_ms - FIFTEEN_MIN
            else:
                event_start_ms = last_check_ms - THIRTY_SEC
            event_stop_ms = utc_time_ms
            try:
                cam_info = request('{proxy}/service/cameras/{cam_id}'
                                   .format(proxy=self.orchid.proxy, cam_id=camera_id))
                if 'primaryStream' not in cam_info:
                    continue
                primary_stream_id = cam_info['primaryStream']['id']
                motion_events = request('{proxy}/service/events/camera-stream'
                                        '?start={start}&stop={stop}&id={ps_id}&eventType=IPC_MOTION_EVENT'
                                        .format(proxy=self.orchid.proxy, start=event_start_ms,
                                                stop=event_stop_ms, ps_id=primary_stream_id))
            except urllib.error.HTTPError as http_err:
                log.error('Failed to obtain motion events for %s camera ID=%d (HTTP %d)',
                          self.orchid.name, camera_id, http_err.code)
                continue
            except OSError as os_err:
                log.error('Unable to request motion events for %s camera ID=%d: %s',
                          self.orchid.name, camera_id, os_err)
                continue
            # filter discovered events and send alerts if necessary
            suppressed_alerts = 0
            for event in motion_events['events']:
                utc_time_ms = int(utc_datetime(epoch=True) * 1000)
                ongoing_event = event['start'] == event['stop']
                if ongoing_event:
                    event_duration_ms = utc_time_ms - event['start']
                else:
                    event_duration_ms = event['stop'] - event['start']
                if event_duration_ms < min_motion_ms:
                    continue
                try: # this entire try block must be serialized
                    db_lock.acquire()
                    existing_event = db.execute('''
                    SELECT orchid_uuid, camera_id, timestamp
                     FROM alert
                    WHERE orchid_uuid=? AND camera_id=? AND timestamp=?
                      ''', (self.orchid.uuid,
                          camera_id, event['start'])).fetchone()
                    if existing_event:
                        continue
                    # check if email needs to be throttled
                    if email_deadband_ms > 0:
                        last_email_ms, = db.execute('''
                        SELECT last_email
                          FROM orchid_camera
                         WHERE orchid_uuid=? AND camera_id=?
                         ''', (self.orchid.uuid, camera_id)).fetchone()
                        if utc_time_ms - last_email_ms < email_deadband_ms:
                            suppressed_alerts += 1
                            continue
                        with db:
                            # further emails will be suppressed for <email_deadband_sec> time
                            db.execute('''
                            UPDATE orchid_camera
                               SET last_email=?
                             WHERE orchid_uuid=? AND camera_id=?
                             ''', (utc_time_ms, self.orchid.uuid, camera_id))
                    with db:
                        # mark this alert as handled (no dupes)
                        db.execute('''
                        INSERT INTO alert(orchid_uuid, camera_id, timestamp)
                        VALUES(?,?,?)
                        ''', (self.orchid.uuid, camera_id, event['start']))
                finally: db_lock.release()
                # cap the export duration if necessary
                if event_duration_ms > max_export_ms or ongoing_event:
                    event['stop'] = event['start'] + max_export_ms
                export_info = {
                    'camera': {
                        'id': camera_id,
                        'name': cam_info['name'],
                        'primaryStreamId': primary_stream_id
                    },
                    'event': event
                }
                # export and email events asynchronously
                job = threading.Thread(target=self._export_and_email, args=(export_info,))
                job.start()
                jobs.append(job)
                log.debug('[%s] started %s',
                          self.orchid.name, job.name)
            if suppressed_alerts > 0:
                log.debug('%d alerts suppressed for %s camera %s',
                          suppressed_alerts, self.orchid.name, cam_info['name'])
        for job in jobs:
            job.join()
            log.debug('[%s] stopped %s',
                      self.orchid.name, job.name)

    def _export_and_email(self, export_info):
        """Internal: export video and email a motion alert to recipients.
        """
        export_url = ('{proxy}/service/streams/{ps_id}/export?start={start}&stop={stop}&format=mov'
                      .format(proxy=self.orchid.proxy, ps_id=export_info['camera']['primaryStreamId'],
                              start=export_info['event']['start'], stop=export_info['event']['stop']))
        file_name = '{}.mov'.format(hashlib.sha1(bytes(export_url, 'utf-8')).hexdigest())
        file_path = Path(gettempdir()) / file_name
        try:
            data = request(export_url, respond_json=False)
        except urllib.error.HTTPError as http_err:
            log.error('Failed exporting video for %s camera %s (HTTP %d)',
                      self.orchid.name, export_info['camera']['name'], http_err.code)
            return
        except OSError as os_err:
            log.error('Unable to export video for %s camera %s: %s',
                      self.orchid.name, export_info['camera']['name'], os_err)
            return
        try:
            file_path.write_bytes(data)
            html = email_template.format(
                    orchid_name=self.orchid.name,
                    camera_name=export_info['camera']['name'],
                    event_time=self.orchid.local_datetime(
                        timestamp_ms=export_info['event']['start']
                        ).strftime('%D - %I:%M %p'), # MM/DD/YY - HH:MM AM/PM
                    tz=self.orchid.tz['name'])
            self._email_alert(html=html, attachment=file_path)
        except SMTPException as smtp_exc:
            err_msg = ('Failed to send email for {server} camera {camera}'
                       .format(server=self.orchid.name, camera=export_info['camera']['name']))
            if isinstance(smtp_exc, SMTPServerDisconnected):
                log.warning('%s: disconnected from %s', err_msg, config['smtp']['server'])
            else:
                log.error('%s: %s', err_msg, smtp_exc)
        finally:
            if file_path.exists():
                file_path.unlink()

    def _email_alert(self, html, attachment):
        """Internal

        Send recipients an HTML formatted email with video attachment.
        """
        smtp_config = config['smtp']
        sender = smtp_config.get('sender', smtp_config['user'])
        # compose message
        msg = MIMEMultipart()
        msg['Subject'] = 'Motion Alert for {}'.format(self.orchid.name)
        msg['From'] = sender
        msg['To'] = ','.join(self.recipients)
        msg.attach(MIMEText(html, 'html'))
        content = MIMEBase('application', 'octet-stream')
        content.set_payload(attachment.read_bytes())
        encoders.encode_base64(content)
        content.add_header('Content-Disposition',
                'attachment', filename=attachment.name)
        msg.attach(content)
        # email message
        smtp = None
        try:
            smtp = SMTP(smtp_config['server'],
                        timeout=smtp_config.getint('connect_timeout_sec', 60))
            smtp.login(smtp_config['user'], smtp_config['password'])
            refused_recipients = smtp.sendmail(sender, self.recipients, msg.as_string())
            # If refused_recipients is empty, it only means the SMTP server accepted all recipients
            # and attempted delivery to their indox. It does not indicate delivery was successful.
            if refused_recipients:
                for recipient in refused_recipients:
                    smtp_code = refused_recipients[recipient][0]
                    reason = refused_recipients[recipient][1].decode('utf-8')
                    log.warning('%s refused to send %s an alert: %s (SMTP %d)',
                                smtp_config['server'], recipient, reason, smtp_code)
        except SMTPException:
            raise # re-raise exception to caller
        finally:
            # only close if actually connected
            if smtp and smtp.noop()[0] == 250:
                smtp.quit()

    def _load_schedules(self):
        """Internal: load Orchid's monitoring schedules.
        """
        schedules = db.execute('''
        SELECT start, stop, days
          FROM orchid_schedule, schedule
         WHERE orchid_schedule.orchid_uuid=?
           AND orchid_schedule.schedule_id=schedule.id
           ''', (self.orchid.uuid,)).fetchall()
        return [ Schedule(start, stop, days)
                 for start, stop, days in schedules ]

    def _load_cameras(self, registered_cams):
        """Internal: load monitored Orchid cameras.

        Performs automatic cleanup for any camera IDs that are in the
        database but not registered with Orchid.
        """
        # collect what's currently in the database
        cam_query = db.execute('''
        SELECT camera_id FROM orchid_camera WHERE orchid_uuid=?
        ''', (self.orchid.uuid,)).fetchall()
        db_cams = set(camera_id for camera_id, in cam_query)
        # filter out any cameras that have been removed
        diff = db_cams.difference(cam['id'] for cam in registered_cams)
        # if nothing changed return current database entries
        if not diff:
            return db_cams
        # otherwise remove the stale cameras and return the update
        log.warning('Detected camera IDs not registered with %s: '
                    '%s (removing from database)', self.orchid.name, diff)
        try:
            db_lock.acquire()
            with db:
                db.execute('''
                DELETE FROM orchid_camera
                 WHERE orchid_uuid=? AND camera_id IN ({})
                   '''.format(','.join(['?'] * len(diff))),
                   (self.orchid.uuid,) + tuple(diff))
        finally:
            db_lock.release()
        return db_cams - diff

############################################## utility functions

def request(url, respond_json=True):
    """Perform an HTTP resource request.

    If 'respond_json' is True, return the response as a utf-8
    JSON object, otherwise return the raw response.

    RCT 6/23/2021 - Added code to ignore SSL errors.  We need
                    this for /most/ Cloud and Alto instances
                    where the script ends up trying to access
                    https://localhost.
                    https://stackoverflow.com/questions/36600583/python-3-urllib-ignore-ssl-certificate-verification
    """
    ctx = ssl.create_default_context()
    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE

    response = urllib.request.urlopen(
            urllib.request.Request(url, headers=header),
            timeout=config['http'].getint('timeout_sec', 180),
            context=ctx).read()
    if respond_json:
        return json.loads(response.decode('utf-8'))
    return response

def utc_datetime(timestamp=None, epoch=False):
    """Obtain a UTC datetime.

    If 'timestamp' (in seconds) is specified, return the datetime
    associated to that timestamp.

    If 'epoch' is True return the current UTC datetime as a POSIX
    epoch timestamp, otherwise return the current UTC datetime.
    """
    if timestamp:
        return datetime.fromtimestamp(timestamp, timezone.utc)
    utc_time = datetime.now(timezone.utc)
    if epoch:
        return utc_time.timestamp()
    return utc_time

def b64encode_auth():
    """Base64 encode Fusion's authorization credentials.

    Format is for a basic authorization header (user:password).
    """
    user = config['fusion']['user']
    password = config['fusion']['password']
    auth_bstring = bytes('{}:{}'.format(user, password), 'utf-8')
    return base64.b64encode(auth_bstring).decode('utf-8')

def get_fusion_uri():
    """Construct the Fusion URI from Fusion's properties file.
    """
    file_path = config['fusion']['properties_path']
    props = ConfigParser()
    props.read_string('[fusion]\n' + Path(file_path).read_text())
    fusion_props = props['fusion']
    return ('{protocol}://127.0.0.1:{port}'
            .format(protocol=fusion_props['listening.protocol'],
                    port=fusion_props['listening.port']))

################################################################

def poll_orchid(orchid_info, recipients):
    motion_alerter = MotionAlerter(orchid_info, recipients)
    if motion_alerter.in_schedule():
        log.debug('Polling %s', orchid_info['name'])
        motion_alerter.alert_if_motion_detected()
        log.debug('Finished polling %s', orchid_info['name'])
    else:
        log.debug('%s is not within an active polling schedule',
                  orchid_info['name'])

def run_service():
    """Run the motion alerter service.

    Performs automatic cleanup for Orchid UUIDs that are in the
    database but not registered with Fusion.
    """

    orchid_query = '''
    SELECT uuid, recipients
      FROM orchid LEFT JOIN recipient_group
        ON orchid.recipient_group_id=recipient_group.id
     '''

    polling_sec = config['polling'].getint('interval_sec', 60)

    while True:
        stale_orchids = []
        next_run = utc_datetime(epoch=True) + polling_sec
        for orchid_uuid, recipients in db.execute(orchid_query).fetchall():
            try:
                orchid_info = request('{uri}/service/orchids/{uuid}'
                                      .format(uri=fusion_uri, uuid=orchid_uuid))
            except urllib.error.HTTPError as http_err:
                if http_err.code == 404:
                    stale_orchids.append(orchid_uuid)
                    log.warning('Detected Orchid UUID not registered with Fusion: %s '
                                '(flagged to be removed from database)', orchid_uuid)
                else:
                    log.error('Failed to obtain information on Orchid UUID=%s '
                              '(HTTP %d), skipping it this run', orchid_uuid, http_err.code)
                continue
            except OSError as os_err:
                log.warning('Issue reaching %s -> %s', fusion_uri, os_err)
                break # break loop and sleep if fusion is unreachable
            if not orchid_info['isAvailable']:
                log.warning('Server %s is unavailable, '
                            'skipping it this run', orchid_info['name'])
                continue
            if not recipients:
                continue
            # 'poll_orchid' workers run in the background until finished,
            # or main thread recieves a SIGTERM (or ctrl-c).
            threading.Thread(target=poll_orchid, args=(orchid_info, recipients)).start()
        if stale_orchids:
            try:
                db_lock.acquire()
                with db:
                    db.execute('''
                    DELETE FROM orchid
                     WHERE uuid in ({})
                     '''.format(','.join(['?'] * len(stale_orchids))),
                     stale_orchids)
            finally: db_lock.release()
        sleep_seconds = next_run - utc_datetime(epoch=True)
        if sleep_seconds > 0:
            log.debug('Sleeping for %d seconds', sleep_seconds)
            time.sleep(sleep_seconds)

def sigterm_handler(signum, frame):
    clean_exit()

def clean_exit():
    # make sure all threads are joined before closing db
    for thread in threading.enumerate():
        if thread is not threading.main_thread():
            thread.join()
    db.close()
    log.info('Stopped motion alerts')
    sys.exit(0)

if __name__ == '__main__':
    config = ConfigParser()
    config.read('config.ini')

    # misc globals
    fusion_uri = get_fusion_uri()
    header = { 'Authorization': 'Basic ' + b64encode_auth() }
    email_template = Path('email_body.html.template').read_text()

    # db setup/connection
    db_lock = threading.RLock()
    db_path = config['database']['path']
    if not os.path.exists(db_path):
        db_schema.create(db_path)
        log.info('Created database at %s (schema version %d)',
                 db_path, db_schema.version)
    # sharing db connection between threads, use write-locks
    db = sqlite3.connect(db_path, check_same_thread=False, timeout=15)
    db.execute('PRAGMA foreign_keys = ON')
    db.execute('PRAGMA journal_mode = WAL')

    signal(SIGTERM, sigterm_handler)
    try:
        log.info('Starting motion alerts')
        run_service()
    except KeyboardInterrupt:
        print('\rcleaning up')
        clean_exit()
