#!/usr/bin/env python3

import os

from routes import RecipientGroup, Schedule, AlertConfiguration
from urllib.parse import urlparse
from sqlite3 import IntegrityError

def send_response(body, code, content_type='application/json'):
    """Send CGI response with content.
    """
    print('Status:', code)
    print('Content-Type:', content_type)
    print()
    print(body)

def parse_route(route):
    """Parse endpoint routes into respective parts.

    '/endpoint'   -> ('endpoint', None)
    '/endpoint/1' -> ('endpoint', '1')

    '/endpoint' and '/endpoint/' return the same, likewise
    for '/endpoint/1/'.

    If we encounter '/endpoint/1/whatever', we only return
    ('endpoint', '1'), for now.
    """
    # filter removes any empty strings created by split
    endpoint, *parts = filter(None,
            urlparse(route).path[1:].split('/'))
    if not parts:
        return (endpoint, None)
    return (endpoint, parts[0])

def handler(obj, method, resource):
    """Generic HTTP route handler.

    We make the assumption that 'obj' is a concrete subclass
    of routes.Route and returns a (response body, status code)
    tuple.
    """
    try:
        if method == 'GET':
            return obj.get(resource)
        if method == 'POST':
            return obj.post()
        if method == 'PUT' and resource:
            return obj.put(resource)
        if method == 'DELETE' and resource:
            return obj.delete(resource)
    except (ValueError, IntegrityError) as bad_request:
        return ('{}'.format(bad_request), 400)
    except Exception as server_error:
        return ('{}'.format(server_error), 500)
    return ('Not Implemented', 501)

if __name__ == '__main__':
    dispatch = {
        'recipientGroups': RecipientGroup(),
        'schedules': Schedule(),
        'alertConfigurations': AlertConfiguration()
    }
    method = os.environ['REQUEST_METHOD']
    endpoint, resource = parse_route(os.environ['PATH_INFO'])
    if endpoint not in dispatch:
        send_response('Not Found', 404)
    response, code = handler(dispatch[endpoint], method=method,
                             resource=resource)
    send_response(response, code)
